/*
 * Copyright © 2007 Global Locate.
 *
 * Written by Ken Yale.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "agps.h"
#include "sv_bar.h"

// #define OPT_TEST_AVERAGE
#undef  OPT_TEST_AVERAGE

#undef  DEBUG_PAGE_MAP

#ifdef  OPT_TEST_AVERAGE
#define DEBUG_PAGE_MAP
#endif  

static GtkAdjustment* adj_center;
static GtkButton*     button_style;
static GtkWidget*     scale_center;

static int adj_upper = 0;
static int adj_lower = 1;

static void haveNewPosition(int n);

//------------------------------------------------------------------------------------
//
//              page_map.cpp
//
//      Plot the GPS positions in a window.
//      Follow the first, last, and average position.
//      Valid for -88 to +88 degrees.
//
//------------------------------------------------------------------------------------

class lla;
const double PI = 3.1415926535;
static double pixels_per_meter;
static void MAP_draw_position(lla* p, bool suppress = true);
static void MAP_recompute_xy_pixels(const char* msg);
void page_MAP_init();
static void MAP_draw_avg_circle(lla* p);
GtkLabel* map_label;
extern int fix_good;

static GdkGC* blue_gc = 0;
static GdkGC* grey_gc = 0;

static GtkDrawingArea* MAP;
static GdkPixmap*      MAP_PIX = 0;
bool mph = false;
static PangoLayout* MAP_text;

//      KLUDGE for A3 or larger fonts only:
// bool large_font  = true;
// bool a3_hardware = true;
#define large_font  1
#define a3_hardware 1

extern GdkGC* alloc_color(GdkPixmap* pm, const char* color);
extern void notebook_expose(const char* pageName);
//static double meter_scale();
void scale_init(GtkScale* scale);
static void MAP_draw_all_points();
void page_MAP_alloc_colors();
static void MAP_redraw();
void page_MAP_have_lla(double lon, double lat, int alt, int heading, int speed, int time);
static void MAP_draw_scale(double mult, int meters, int nDividers);

static bool map_have_new_pos = false;

static int lla_x0;
static int lla_y0;

#define MAX_X   ((1 << 15) - 1)
#define MIN_X   (-MAX_X)
#define MAX_Y   MAX_X
#define MIN_Y   MIN_X
#define SCREEN_WIDTH   476
#define SCREEN_HEIGHT  400


//------------------------------------------------------------------------------------
//
//      Class lla
//
//  Save the longitude and latitude.
//  (the altitude, heading, speed, time are also saved).
//  Convert LLA one time to meters.
//  Convert to an integer pixel offset from the previous point when
//  the map scale changes.
//
//------------------------------------------------------------------------------------

class lla
{
public:
    lla() : time(0)
    {
    };

    lla(double lon_, double lat_, int alt_, int head, int sp_, unsigned long t_)
        :
                lon(lon_),
                lat(lat_),
                alt(alt_),
                heading(head),
                speed(sp_),
                time(t_)
    {
    };

    void init(double lon_, double lat_, int alt_, int head, int sp_, unsigned long t_)
    {
        lon = lon_;
        lat = lat_;
        alt = alt_;
        heading = head;
        speed = sp_;
        time = t_;
        compute_xy_meters();
        last = *this;
    };

    ~lla()
    {
    };

    void compute_xy_meters(void);
    void compute_xy_meters_abs(lla* p);
    void compute_xy_pixels(const char* msg = 0);

    double lon;
    double lat;
    double delta_lon;
    double delta_lat;

    double meters_x;    // relative to previous point.
    double meters_y;    // relative to previous point.

    int pixels_x;       // relative to previous point.
    int pixels_y;

    int    alt;         // meters
    int    heading;     // 0=N  90=E  180=S  270=W
    int    speed;       // km/h
    unsigned long time; // ms into the GPS week.

    static void restart(void)
    {
        prev = 0;
    }
    static lla first;

    static lla* lastPoint()
    {
        return &last;
    };

private:
    static lla* prev;
    static lla  last;
};

#ifdef  DEBUG_PAGE_MAP
// #define DEBUG_F(f)     printf("    " #f " %f\n", f)
#define DEBUG_F(f)
#else
#define DEBUG_F(f)
#endif

enum
{
#ifdef OPT_TEST_AVERAGE
    MAX_LLA = 7
#else
    MAX_LLA = 2000
#endif
};

lla avg_lla;
lla LLA[MAX_LLA];
static int nLLA = 0;
static int indexLLA = 0;

static double sum_lon;
static double sum_lat;
// static int    sum_alt = 0;

lla  lla::last;
lla  lla::first;
lla* lla::prev = 0;

void lla::compute_xy_meters()
{
    if (prev)
    {
        compute_xy_meters_abs(prev);
    }
    else
    {
        compute_xy_meters_abs(this);
        first = *this;
    }
    last = *this;
    prev = this;
}

//------------------------------------------------------------------------------------
//
//      Class MeterScale
//
//  Here are all the choices of what you can select with the +/- buttons.
//
//------------------------------------------------------------------------------------

struct MeterScale
{
    enum
    {
        SCALE_1x1  =  9
    };
    double scale;               // Pixels/meter
    int ruler_meters;
    int ruler_dividers;
    const char* name;
};

MeterScale scales[] =
{
    { 50, 2, 2, "x50" },           // 0
    { 32, 5, 5, "x32" },
    { 20, 5, 5, "x20" },
    { 16, 10, 5, "x16" },

    { 10, 10, 5, "x10" },           // 4
    {  8, 20, 4, "x8"  },

    {  5, 20, 4, "x5"  },           // 6
    {  4, 20, 4, "x4"  },
    {  2, 50, 5, "x2"  },

    { 1.0,  100, 5, "x1"  },           // 9

    { 0.5,  200, 4, "/2"  },           // 10
    {  .2,  500, 5, "/5"  },
    {  .1,  1000, 5, "/10" },

    { .05,  2000,  4, "/20"  },        // 13
    { .02,  5000,  5, "/50"  },
    { .01,  10000, 5, "/100" },

    { .005,   20000,  5, "/200" },     // 16
    { .002,   50000,  5, "/500" },
    { .001,   100000, 5, "/1k"  },

    { .0005,  200000,  4, "/2k"  },    // 19
    { .0002,  500000,  5, "/5k"  }
};
#define MAX_SCALES (sizeof(scales) / sizeof(*scales))

static int scale_index = MeterScale::SCALE_1x1;

//-------------------------------------------------------------------------------------
//
//              testAverage()
//
//-------------------------------------------------------------------------------------

#ifdef OPT_TEST_AVERAGE
static bool
testAverage()
{
    static bool average_done = false;
    static bool been_here    = false;

    if (been_here) return average_done;
    been_here = true;

    page_MAP_have_lla(0.000010, 0.000010, 50, 12, 13, 12351);
    page_MAP_have_lla(0.000020, 0.000020, 50, 12, 13, 12352);
    page_MAP_have_lla(0.000030, 0.000030, 50, 12, 13, 12353);   // Average/5
    page_MAP_have_lla(0.000040, 0.000040, 50, 12, 13, 12354);
    page_MAP_have_lla(0.000050, 0.000050, 50, 12, 13, 12355);

    page_MAP_have_lla(0.000210, 0.000290, 50, 12, 13, 12356);
    page_MAP_have_lla(0.000220, 0.000280, 50, 12, 13, 12357);
    page_MAP_have_lla(0.000250, 0.000250, 50, 12, 13, 12358);   // Average/5
    page_MAP_have_lla(0.000280, 0.000220, 50, 12, 13, 12359);
    page_MAP_have_lla(0.000290, 0.000210, 50, 12, 13, 12360);

    average_done = true;
    return average_done;
}

#define TEST_AVERAGE()  testAverage()
// #define TEST_AVERAGE()
#else
#define TEST_AVERAGE()     false
#endif

//------------------------------------------------------------------------------
//
//      IMPORTANT NOTE:
//
//      This code is intended ONLY to show relative positions
//      of lat/lon on an x,y pixmap to look for GPS problems.
//      It is NOT to be used for any accurate mapping
//      to reference points, reference to any geoid, or any
//      form of navigation.
//
//------------------------------------------------------------------------------

void lla::compute_xy_meters_abs(lla* p)
{
#ifdef  DEBUG_PAGE_MAP
    printf("compute_xy_meters[%f %f]-[%f %f]", lon, lat, p->lon, p->lat);
#endif

    double average_lat = (p->lat + lat)/2.0;            DEBUG_F(average_lat);
    double rad_lat     = average_lat * PI / 180.0;      DEBUG_F(rad_lat);

    double m = 111132.09;                               DEBUG_F(m);
    m -= 566.05 * cos(2.0 * rad_lat);                   DEBUG_F(m);
    m += 1.2*cos(4*rad_lat);                            DEBUG_F(m);

    delta_lat = lat - p->lat;                           DEBUG_F(delta_lat);
    meters_y = delta_lat * m;                           DEBUG_F(meters_y);

    double pp = 111415.13 * cos(rad_lat) - 94.55*cos(3*rad_lat); DEBUG_F(pp);
    delta_lon = lon - p->lon;                           DEBUG_F(delta_lon);
    if (delta_lon <= -180.0)
    {
        delta_lon += 360.0;                             // printf("#%f#", delta_lon);
    }
    else if (delta_lon > 180.0)
    {
        delta_lon -= 360.0;                             // printf("$%f$", delta_lon);
    }
    meters_x = delta_lon * pp;                          DEBUG_F(meters_x);

#ifdef  DEBUG_PAGE_MAP
    printf("= [%f %f]  ", delta_lon, delta_lat);
    printf("= {%f %f}m\n", meters_x,  meters_y );
#endif
}

void
lla::compute_xy_pixels(const char* msg)
{
    // The negative sign on the Y represents the GTK+ coordinate system
    // where X increases to the right and Y increases as you go lower
    // on the screen.
    pixels_x =  static_cast<int> (pixels_per_meter * meters_x);
    pixels_y = -static_cast<int> (pixels_per_meter * meters_y);
#ifdef  DEBUG_PAGE_MAP
    if (msg)
    {
        printf("compute_xy_pixels(%s) (%d %d) = %f x {%f %f}\n",
                msg,
                pixels_x, pixels_y, pixels_per_meter,
                meters_x, meters_y);
    }
#endif
}

static lla*
addLLA()
{
    lla* result = LLA + indexLLA;
    if (++indexLLA >= MAX_LLA) indexLLA = 0;

    if (++nLLA > MAX_LLA)
    {
        nLLA     = MAX_LLA;             // Limit ourselves to our LLA array size.
        sum_lon -= result->lon;         // The average is over what we have in
        sum_lat -= result->lat;         // the array, so remove older items from
        // sum_alt -= result->alt;      // the running average.
#ifdef DEBUG_PAGE_MAP
        printf("SUB [%f %f]\n", result->lon, result->lat);
#endif
        //  Since we're about to wipe out this item, make sure the *next*
        //  item is still tracked relative to the initial position.
        //  To keep it simple, reassign the initial position to be the next
        //  position:

        lla* next = LLA + indexLLA;
        next->meters_x = 0;
        next->meters_y = 0;
        next->delta_lon = 0;
        next->delta_lat = 0;
        lla::first = *next;
    }

#ifdef DEBUG_PAGE_MAP
    printf("have_LLA %d --> %d %d\n", old_n, nLLA, fix_good);
#endif
    haveNewPosition(fix_good);

    return result;
}

//------------------------------------------------------------------------------
//
//      lla iterator
//
//  OpenMoko apparently lacks <deque> and <algorithm>, so here is a hacked-version
//  of some iterators for the lla{} vector.
//
//------------------------------------------------------------------------------

static lla* pEnd;
static lla* pCur;
static lla* pWrap = LLA + MAX_LLA;

lla* lla_next()
{
    if (++pCur >= pWrap)
    {
        pCur = LLA;
    }
    return pCur;
}

lla* lla_pointN(int n)
{
    int i = n;
    if (nLLA >= MAX_LLA) i = (indexLLA + i) % MAX_LLA;

    lla* r = LLA + i;
#ifdef  DEBUG_PAGE_MAP
    printf("lla_pointN(%d) LLA[%d]=[%f %f]\n", n, i, r->lon, r->lat);
#endif
    return r;
}

bool lla_last()
{
    return (pCur == pEnd);
}

lla* lla_begin()
{
    if (!nLLA) return 0;
    pCur = LLA;
    pEnd = LLA + indexLLA;
    if (nLLA >= MAX_LLA)
    {
        pCur = LLA + indexLLA;
        pEnd = pCur;
        // if (++pCur >= pWrap)
        if (pCur >= pWrap)
        {
            pCur = LLA;
        }
    }
    return pCur;
}

//------------------------------------------------------------------------------
//
//      Class TrackStyle
//
//  Follow the first, average, or last position.
//
//------------------------------------------------------------------------------

class TrackStyle
{
public:
    enum TrackStyleEnum
    {
        POINT_0 = 0,
        AVERAGE,
        POINT_N,
        POINT_LAST,
        STYLE_MAX
    };

    static bool increment(void);
    static const char* getName(char* buf);
    static const char* getLabel(void);
    static lla* getReference(void);
    static lla* getReference(TrackStyleEnum style);
    static TrackStyleEnum getStyle(void);
    static bool NewPositionCenter(int i);

    TrackStyleEnum style;
    const char*    label;
    const char*    name;

private:
    static TrackStyle styles[STYLE_MAX];
    static int index;
    static int point_n;
    static lla* reference_position;
};

static lla lla_n;           // the Nth point.

lla*
TrackStyle::getReference(TrackStyleEnum i)
{
    lla* r = 0;
    switch (i)
    {
    default:
    case POINT_0:
        break;
    case AVERAGE:
        r = &avg_lla;
        break;
    case POINT_N:
        // Convert the intuitive, user-based index that starts at 1.
        // into internal point index "point_n" which is 0-based
        lla_n = *lla_pointN(point_n-1);
        r     = &lla_n;
        break;
    case POINT_LAST:
        r = lla::lastPoint();
        break;
    }
    return r;
}

lla*
TrackStyle::getReference(void)
{
    return getReference(styles[index].style);
}

int  TrackStyle::index = TrackStyle::AVERAGE;   // Default tracking style
lla* TrackStyle::reference_position = 0;
int TrackStyle::point_n;

TrackStyle::TrackStyleEnum TrackStyle::getStyle()
{
    return styles[index].style;
}

TrackStyle TrackStyle::styles[TrackStyle::STYLE_MAX] =
{
    {TrackStyle::POINT_0,    "First","First Fix"},
    {TrackStyle::AVERAGE,    "Avg",  "Average"  },
    {TrackStyle::POINT_N,    "Fix",  "Fix #%d"   },
    {TrackStyle::POINT_LAST, "Last", "Last Fix" }
};

static void
haveNewPosition(int n)
{
    adj_upper = n;
    const char* msg = 0;
    if (TrackStyle::getStyle() == TrackStyle::POINT_LAST)
    {
        msg = "value";
    }
    g_object_set(G_OBJECT(adj_center),
                "upper", (gdouble) n,
                msg,     (gdouble) n,
                NULL);
    gtk_adjustment_changed(adj_center);
    if (msg)
    {
        gtk_adjustment_value_changed(adj_center);
    }
    if (n > MAX_LLA)
    {
        //------------------------------------------------
        //
        // The LLA array is full, and we've wrapped around.
        // So the user will see numbers larger, but internally,
        // we have to map them as shown:
        //       AVERAGE          FIRST         LAST
        // user: n-MAX_LLA-1      n-MAX_LLA     n
        // LLA[] 0                1             MAX_LLA
        //
        //------------------------------------------------

        adj_lower = n - MAX_LLA;
        g_object_set(G_OBJECT(adj_center), "lower", (gdouble) adj_lower, NULL);
    }
}

const char* TrackStyle::getLabel(void)
{
    return styles[index].label;
}

const char* TrackStyle::getName(char* buf)
{
    sprintf(buf, styles[index].name, point_n);
    return buf;
}

bool TrackStyle::increment(void)
{
    if (++index >= STYLE_MAX)
    {
        index = POINT_0;
    }
    return (index == POINT_N);
}

bool
TrackStyle::NewPositionCenter(int adj)
{
    bool rc = true;

    if (adj == point_n)
    {
        // No change, so this is a self-induced jump!
        printf("NewPositionCenter(%d) no change\n", adj);
        return rc;
    }

    lla* p = lla_pointN(adj - 1);

    //  Moved to the 1st position, so just go there directly.
    if (lla_begin() == p)
    {
        point_n = adj;
        printf("NewPositionCenter(%d) first\n", adj);
        return rc;
    }

    int new_adj = adj;
    if (adj < point_n)   // go backwards, but don't go past the start.
    {
        int lower_limit = fix_good - MAX_LLA;
        if (lower_limit < 0) lower_limit = 1;

        // Since we're going backwards, look in the *NEXT* position
        // to see if we have a position change.
        p = lla_pointN(new_adj);

        while (p->pixels_x == 0 && p->pixels_y == 0)
        {
            --new_adj;
            if (new_adj < lower_limit)
            {
                printf("NewPositionCenter(%d) first %d\n", adj, new_adj);
                rc = false;
            }
            p = lla_pointN(new_adj);
        }
    }
    else                // go forwards, but don't go past the end.
    {
        while (p->pixels_x == 0 && p->pixels_y == 0)
        {
            ++new_adj;
            if (new_adj > fix_good)
            {
                printf("NewPositionCenter(%d) last %d\n", adj, new_adj);
                return false;
            }
            p = lla_pointN(new_adj - 1);
        }
    }

    point_n = new_adj;
    if (new_adj != adj) rc = false;

    if (!rc)
    {
        //      Since we're skipping points, don't
        //      update now:  expect to be called to do the update.
        g_object_set(G_OBJECT(adj_center), "value", (gdouble) new_adj, NULL);
        gtk_adjustment_value_changed(adj_center);
        printf("NewPositionCenter(%d) skip to %d\n", adj, new_adj);
    }
    return rc;
}

static int pixel_offset_x;
static int pixel_offset_y;

static void
centerOfWindow(int& x, int& y)
{
    const int screen_center_x = 476/2;
    const int screen_center_y = 400/2;

    x = screen_center_x;
    y = screen_center_y;
}

const char*
formatLL(double lon, double lat)
{
    char ew = 'E';
    char ns = 'N';

    if (lon < 0.0) ew = 'W';
    if (lat < 0.0) ns = 'S';

    static char buf[64];
    sprintf(buf, "%f %c  %f %c", lon, ew, lat, ns);
    return buf;
}

//-----------------------------------------------------------------------------------
//
//              computePixelBaseline()
//
//      The pixel baseline is added to each pixels_x,pixels_y of the lla{}.
//      It is:
//      a)  position[0] (or 0,0) + "screen center" 
//          When the center is the first position fix.
//      b)  position[0]-avg + "screen center" 
//          When the center is the average position.
//      c)  position[0]-position[LAST] + "screen center" 
//          When the center is the last position.
//      d)  position[0]-position[n] + "screen center" 
//          When the center is the position[n].
//
//      Since position[N].meters_x and .pixels_x are computed relative
//      to position[N-1], we need to compute a temporary:
//              temp_lla = position[N] - position[0]
//
//-----------------------------------------------------------------------------------

void
computePixelBaseline()
{
    centerOfWindow(pixel_offset_x, pixel_offset_y);

    lla* r = TrackStyle::getReference();

    //-----------------------------------------------------------------
    //  If we have a relative lla to use, then figure out how far it is
    //  from [0,0].
    //-----------------------------------------------------------------

    if (r)
    {
        char buf0[64];
        const char* n = TrackStyle::getName(buf0);

        r->compute_xy_meters_abs(&lla::first);
        r->compute_xy_pixels(n);
#ifdef  DEBUG_PAGE_MAP
        printf("baseline (%s)-first = (%d %d) - (%d %d) = (%d %d)\n",
                    buf0, 
                    pixel_offset_x, pixel_offset_y,
                    r->pixels_x, r->pixels_y,
                    pixel_offset_x - r->pixels_x,
                    pixel_offset_y - r->pixels_y);
#endif
        pixel_offset_x -= r->pixels_x;
        pixel_offset_y -= r->pixels_y;
    }
    else
    {
        // printf("baseline FIRST = (%d %d)\n", pixel_offset_x, pixel_offset_y);
    }
    if (r != &avg_lla)
    {
        // printf("redo(avg) ");
        avg_lla.compute_xy_meters_abs(&lla::first);
        avg_lla.compute_xy_pixels("redo(avg)");
    }
}

//-----------------------------------------------------------------------------------
//
//      set_fix_label()
//
//-----------------------------------------------------------------------------------

static void
set_fix_label()
{
    lla* p = TrackStyle::getReference();
    if (!p) p = &lla::first;

    char buf0[64];
    char buf1[128];
    sprintf(buf1, "%s %s #%d",
                        TrackStyle::getName(buf0),
                        formatLL(p->lon, p->lat),
                        fix_good);
    gtk_label_set_text(map_label, buf1);
}

//-----------------------------------------------------------------------------------
//
//      page_MAP_have_lla()
//
//  nmea.cpp sends us positions when they arrive from the NMEA input.
//  Create an lla & save it.
//
//-----------------------------------------------------------------------------------

void
page_MAP_have_lla(      double lon, double lat,
                        int alt,
                        int heading, int speed, int time)
{
    if (TEST_AVERAGE()) return;

#ifdef DEBUG_PAGE_MAP
    printf("page_MAP_have_lla [%f %f]\n", lon, lat);
#endif

    lla* p = addLLA();
    p->init(lon, lat, alt, heading, speed, time);

    bool allowSuppress = true;
    if (nLLA == 1)
    {
        allowSuppress = false;
        sum_lon = lon;
        sum_lat = lat;
    }
    else
    {
        sum_lon += p->lon;
        sum_lat += p->lat;
    }

#ifdef DEBUG_PAGE_MAP
    printf("ADD [%f %f] = \n", p->lon, p->lat);
    printf("sum [%f %f] / %d\n", sum_lon, sum_lat, nLLA);
#endif

    // Compute the average position:
    avg_lla.lon = sum_lon / (double) nLLA;
    if (avg_lla.lon > 180.0)        avg_lla.lon -= 360.0;
    else if (avg_lla.lon <= -180.0) avg_lla.lon += 360.0;

    avg_lla.lat =  sum_lat           / (double) nLLA;
    // avg_lla.alt = (sum_alt + nLLA/2) /          nLLA;

#ifdef DEBUG_PAGE_MAP
    printf("avg [%f %f]\n", avg_lla.lon, avg_lla.lat);
#endif

    map_have_new_pos = true;

    //  Only compute pixel baseline if:
    //  ...   mult changes
    //  ===>  or if average changes (when centering on the       average)  <===
    //  ===>  or if newer position  (when centering on the last position)  <===
    //  ===>  or if newer position  (when we have filled the LLA array  )  <===
    computePixelBaseline();
    p->compute_xy_pixels("new lla");

    if (MAP_PIX)
    {
        MAP_draw_avg_circle(&avg_lla);          // average is a small circle
        MAP_draw_position(p, allowSuppress);
        set_fix_label();
    }
}

//------------------------------------------------------------------------------------
//
//      MAP_self_test
//
//  Create dummy points to exercise the plotting algorithms
//
//------------------------------------------------------------------------------------

static const char*
MAP_self_test()
{
    const char* msg = 0;

    // Stub for now:  add some data points!
    static bool been_here = false;
    if (!been_here)
    {
        been_here = true;

#if 0   // PASS
        msg = "2-point test case";
        page_MAP_have_lla(-1.0, -1.0, 50, 0, 0, 126000);
        page_MAP_have_lla( 1.0,  1.0, 50, 0, 0, 127000);
#if 0   // PASS
        msg = "3-point test case";
        page_MAP_have_lla( 1.1,  1.0, 50, 0, 0, 127000);
#endif
#endif

#if 0   // PASS
        msg = "Local test case";
        page_MAP_have_lla(-121.000000, 37.000000, 50, 0, 0, 123000);
        page_MAP_have_lla(-121.000120, 37.000000, 50, 0, 0, 124000);
        page_MAP_have_lla(-121.000120, 37.001500, 50, 0, 0, 125000);
        page_MAP_have_lla(-121.000000, 37.001500, 50, 0, 0, 126000);
        page_MAP_have_lla(-121.000000, 37.000200, 50, 0, 0, 127000);
#endif

#if 0   // PASS
        msg = "West test case";
        page_MAP_have_lla( 0.000030, -0.00050, 50, 0, 0, 123000);
        page_MAP_have_lla(-0.000120, -0.00050, 50, 0, 0, 124000);
        page_MAP_have_lla(-0.000120,  0.00050, 50, 0, 0, 125000);
        page_MAP_have_lla( 0.000030,  0.00050, 50, 0, 0, 126000);
        page_MAP_have_lla( 0.000030, -0.00045, 50, 0, 0, 127000);
#endif

#if 0   // PASS        // Test case 1 (wrap around "east pole" 180,0)
        msg = "East #1 test case";
        page_MAP_have_lla(-179.99801, -0.000511, 50, 0, 0, 123000);
        page_MAP_have_lla( 179.99812, -0.000522, 50, 0, 0, 124000);
        // page_MAP_have_lla(-179.99803, -0.000536, 50, 0, 0, 123000);

#if 0
        msg = "East #2 test case";
        page_MAP_have_lla( 179.99813, +0.00053, 50, 0, 0, 125000);
        page_MAP_have_lla(-179.99794, +0.00054, 50, 0, 0, 126000);
        page_MAP_have_lla(-179.99805, -0.00050, 50, 0, 0, 127000);
#endif
#endif

#if 0   // FAIL           // Test case 1 (wrap around "north pole" 0,90)
        msg = "North test case\n";
        page_MAP_have_lla(   0.00101, 89.99500, 50, 0, 0, 123000);
        page_MAP_have_lla(  90.00102, 89.99500, 50, 0, 0, 124000);
        page_MAP_have_lla( 179.00103, 89.99500, 50, 0, 0, 125000);
        page_MAP_have_lla( -90.00104, 89.99500, 50, 0, 0, 126000);
        page_MAP_have_lla(  -0.99905, 89.99450, 50, 0, 0, 127000);
#endif

#if 0   // FAIL           // Test case 1 (wrap around "south pole" 0,-90)
        msg = "South test case\n";
        page_MAP_have_lla(   0.00100, -89.99500, 50, 0, 0, 123000);
        page_MAP_have_lla(  90.00100, -89.99500, 50, 0, 0, 124000);
        page_MAP_have_lla( 179.00100, -89.99500, 50, 0, 0, 125000);
        page_MAP_have_lla( -90.00100, -89.99500, 50, 0, 0, 126000);
        page_MAP_have_lla(  -0.99900, -89.99450, 50, 0, 0, 127000);
#endif

    }
    return msg;
}

//-----------------------------------------------------------------------------------
//
//      cb_change_center()
//
// Adjusting slider to change center position
//
//-----------------------------------------------------------------------------------

static void
cb_change_center(GtkAdjustment* adj)
{
    if (TrackStyle::NewPositionCenter((gint) adj->value))
    {
        MAP_redraw();
    }
}

//-----------------------------------------------------------------------------------
//
//      on_map_units_change()
//
//  Do nothing (maybe self-test)
//  Update the page label.
//
//-----------------------------------------------------------------------------------

static void
on_map_units_change()
{
    // page_MAP_init();
    const char* msg = MAP_self_test();
    if (!msg)
    {
        msg = "km/h";
        if (mph) msg = "mph";
    }
    gtk_label_set_text(map_label, msg);
}

//------------------------------------------------------------------------------
//
//      on_clicked_track_style()
//
//  Walk through the different tracking styles.
//
//------------------------------------------------------------------------------

static void
on_clicked_track_style(GtkWidget* w, gpointer d)
{
    gtk_widget_set_sensitive(GTK_WIDGET(scale_center),
                             TrackStyle::increment()  );
    MAP_recompute_xy_pixels("track style change");
    MAP_redraw();
}

//------------------------------------------------------------------------------
//
//      on_clicked_mph()
//
//  GTK+ callback.
//  Update the system state, and the button.
//
//------------------------------------------------------------------------------

static void
on_clicked_mph(GtkWidget* widget, gpointer data)
{
    mph = !mph;
    gtk_button_set_label(GTK_BUTTON(data), (mph) ? "mph" : "km/h");
    on_map_units_change();
}

static void
MAP_clear()
{
    GtkWidget* w = GTK_WIDGET(MAP);
    gdk_draw_rectangle(MAP_PIX, w->style->white_gc, TRUE,
                       0, 0,
                       w->allocation.width,
                       w->allocation.height);
}

static gboolean
configure_MAP_event(GtkWidget* w, GdkEventConfigure* event)
{
    if (MAP_PIX)
    {
        g_object_unref(MAP_PIX);
    }

    //--------------------------------------------------------------------------
    //          pixmap to make the window easier to draw
    //--------------------------------------------------------------------------
    MAP_PIX = gdk_pixmap_new(w->window,
                            w->allocation.width,
                            w->allocation.height, -1);
    MAP_clear();
    return TRUE;
}

static gboolean
expose_MAP_event(GtkWidget* w, GdkEventExpose* event)
{
    notebook_expose("MP");
    GdkRectangle& a = event->area;
    gdk_draw_drawable(w->window,
                      w->style->fg_gc[GTK_WIDGET_STATE(w)],
                      MAP_PIX, a.x, a.y, a.x, a.y, a.width, a.height);
    return FALSE;
}

static GdkGC* 
getGC(bool used)
{
    GdkGC* gc = grey_gc;
    if (used)
    {
        gc = blue_gc;                   // used SV is colored blue
    }
    if (!gc)
    {
        gc = GTK_WIDGET(MAP)->style->black_gc;
    }
    return gc;
}

void
page_MAP_alloc_colors()
{
    if (!MAP_PIX) return;
    if (!blue_gc) blue_gc = alloc_color(MAP_PIX, "Blue");
    if (!grey_gc) grey_gc = alloc_color(MAP_PIX, "SlateGrey");
}

//------------------------------------------------------------------------------
//
//      MAP_draw_avg_circle
//
//  Draw the little "average" circle at the point.
//
//------------------------------------------------------------------------------

static void
MAP_draw_avg_circle(lla* p)
{
    int rect = 5;

    int x = pixel_offset_x + p->pixels_x;
    int y = pixel_offset_y + p->pixels_y;

    if (x > SCREEN_WIDTH || x < 0 || y > SCREEN_HEIGHT || y < 0)
    {
        // printf("circle [%f %f] {%f %f} (%d %d) OFF SCREEN\n",
        //         p->lon, p->lat,
        //         p->meters_x, p->meters_y,
        //         x, y);
        return;
    }

    //----------------------------------------------
    //          Draw 11x11 circle
    //----------------------------------------------

    int size = 2*rect + 1;
#ifdef  DEBUG_PAGE_MAP
    printf("MAP circle (%d %d)==>(%d %d) size %d\n", x, y, x-rect, y-rect, size);
#endif
    gdk_draw_arc(MAP_PIX, GTK_WIDGET(MAP)->style->black_gc, FALSE,
            x - rect, y - rect, size, size, 0, 360 * DEGREE_64);
}

//-----------------------------------------------------------------------------
//
//              Draw the positions and lines between them
//
//      (if they are not too close)
//
//-----------------------------------------------------------------------------

static int accum_x;
static int accum_y;

static void
MAP_reset_position_accum()
{
    accum_x = 0;
    accum_y = 0;
}

static void
MAP_draw_position(lla* p, bool suppress)
{
    //------------------------------------------------------
    //  Suppress everything if the point is too close to the
    //  previous position.
    //  The distance is calculated as dx*dx + dy*dy, where
    //          dx == p->pixels_x and
    //          dy == p->pixels_y in meter units.
    //  If we drop a point, then accumulate the delta(x,y) for
    //  the next point.
    //------------------------------------------------------

    const int rect = 5;
    const int half_rect_sq = (rect/2) * (rect/2);

    int x = p->pixels_x + accum_x;
    int y = p->pixels_y + accum_y;

#ifdef DEBUG_PAGE_MAP
    printf("MAP draw pos (%d %d)", p->pixels_x, p->pixels_y);
    if (accum_x || accum_y) printf("+(%d %d) ", accum_x, accum_y);
#endif

    if (suppress)
    {
        //      We are allowed to suppress this point if the x,y is
        //      too close to the previous point.
        int dist = x*x + y*y;
        if (dist < half_rect_sq)
        {
#ifdef DEBUG_PAGE_MAP
            if (dist) printf(" = %d  TOO CLOSE: suppress", dist);
            printf("\n");
#endif
            accum_x += p->pixels_x;
            accum_y += p->pixels_y;
            return;
        }
    }

    x += lla_x0;
    y += lla_y0;

    MAP_reset_position_accum();

    //----------------------------------------------
    //     The line and position will be grey if it
    //     exceeds bounds;
    //----------------------------------------------

    GdkGC* gc = getGC(true);
    if (x > MAX_X || x < MIN_X || y > MAX_Y || y < MIN_Y)
    {
        if (x > MAX_X) x = MAX_X;
        if (x < MIN_X) x = MIN_X;
        if (y > MAX_Y) y = MAX_Y;
        if (y < MIN_Y) y = MIN_Y;

#ifdef DEBUG_PAGE_MAP
        printf(" ==> (%d %d) > %d..%d bounds exceeded ", x, y, MIN_X, MAX_Y);
#endif
        gc = getGC(false);
    }

    //----------------------------------------------
    //          Draw the point as a 5x5 rectangle
    //  Don't draw if off the screen.
    //----------------------------------------------

    if (x > SCREEN_WIDTH || x < 0 || y > SCREEN_HEIGHT || y < 0)
    {
#ifdef DEBUG_PAGE_MAP
        printf(" > %d..%d : OFF SCREEN", SCREEN_WIDTH, SCREEN_HEIGHT);
#endif
    }
    else
    {
        gdk_draw_rectangle(MAP_PIX, gc, TRUE, x - rect/2, y - rect/2, rect, rect);
#ifdef DEBUG_PAGE_MAP
        printf(" rect @ [%f %f]", p->lon, p->lat);
#endif
    }

    //----------------------------------------------
    //          Draw the route as a line.
    //  DO draw if out of bounds, but make it grey.
    //----------------------------------------------

    if (suppress)
    {
        gdk_draw_line(MAP_PIX, gc, lla_x0, lla_y0, x, y);
#ifdef DEBUG_PAGE_MAP
        // printf(" line (%d %d)-->(%d %d)", lla_x0, lla_y0, x, y);
#endif
    }
#ifdef DEBUG_PAGE_MAP
    printf("\n");
#endif

    lla_x0 = x;
    lla_y0 = y;
}

//------------------------------------------------------------------------------------
//
//      MAP_draw_all_points()
//
//------------------------------------------------------------------------------------

static void
MAP_draw_all_points()
{
    MeterScale& m = scales[scale_index];
    MAP_draw_scale(m.scale, m.ruler_meters, m.ruler_dividers);

    // printf("MAP_draw_all_points() %d %d\n", indexLLA, nLLA);

    if (!nLLA) return;

    lla_x0 = pixel_offset_x;
    lla_y0 = pixel_offset_y;

    MAP_draw_avg_circle(&avg_lla);       // average circle

    lla* p = lla_begin();
    if (!p) return;

    MAP_reset_position_accum();
#ifdef  DEBUG_PAGE_MAP
    int i = 0;
    printf("%d) ", i);
#endif
    MAP_draw_position(p, false);        // first point is not suppressed.
    do
    {
        p = lla_next();
#ifdef  DEBUG_PAGE_MAP
        printf("%d) ", ++i);
#endif
        MAP_draw_position(p);
    } while (!lla_last());

#if   0                 // DEBUG_PAGE_MAP
    i = 0;
    p = lla_begin();
    do
    {
        printf("%d) 0x%x [%f %f]\n", i++, p, p->lon, p->lat);
        p = lla_next();
    } while (!lla_last());
#endif
}

//-----------------------------------------------------------------------------------
//
//      MAP_draw_scale()
//
//  Draw the meter scale, text, sub-markers,
//  circle, and text on the circle.
//
//-----------------------------------------------------------------------------------

static void
MAP_draw_scale(double mult, int meters, int nDividers)
{
    GdkGC* gc = getGC(false);

    double width = mult * static_cast<double> (meters);
    int w = static_cast<int> (width);

    //----------------------------------------------
    //          Draw the scale
    //----------------------------------------------

    int x0 = 10;                // Arbitrary scale location choice...
    int y0 = 20;

    int x1 = x0 + w;
    int y1 = y0;
    gdk_draw_line(MAP_PIX, gc, x0, y0, x1, y1);
    gdk_draw_line(MAP_PIX, gc, x0, y0, x0, y0 + 8);
    gdk_draw_line(MAP_PIX, gc, x1, y1, x1, y1 + 8);

    //----------------------------------------------
    //          Draw the sub-scale markers
    //----------------------------------------------

    int sub = w/nDividers;
    for (x0 += sub; x0 < x1; x0 += sub)
    {
        gdk_draw_line(MAP_PIX, gc, x0, y0, x0, y0 + 4);
    }

    //----------------------------------------------
    //          Compute and format the scale text
    //----------------------------------------------

    char buf[10];
    if (meters >= 1000) sprintf(buf, "%d km", meters/1000);
    else                sprintf(buf, "%d m",  meters     );

    //----------------------------------------------
    //          Draw the scale text
    //----------------------------------------------

    pango_layout_set_text(MAP_text, buf, -1);
    int h;
    pango_layout_get_pixel_size(MAP_text, 0, &h);
    gdk_draw_layout(MAP_PIX, gc, x1 + 3, y0 - h/2, MAP_text);

    //-------------------------------------------------------------------------
    //      Draw the radius circle and text
    //-------------------------------------------------------------------------

    centerOfWindow(x0, y0);

    int size = 2*w + 1;
    // printf("MAP circle (%d %d)==>(%d %d) size %d\n", x0, y0, x0-w, y0-w, size);
    gdk_draw_arc(MAP_PIX, gc, FALSE,
            x0-w, y0-w, size, size, 0, 360 * DEGREE_64);
    gdk_draw_layout(MAP_PIX, gc, x0 + w + 3, y0 - h/2, MAP_text);

#if 0
    //----------------------------------------------
    //          Draw 15x15 cross
    //----------------------------------------------
    gdk_draw_line(MAP_PIX, gc, x0-7, y0  , x0+7, y0  );
    gdk_draw_line(MAP_PIX, gc, x0  , y0-7, x0  , y0+7);
#else
    //-------------------------------------------------------------------------
    //      Experiment with a NxN grey spot in the center.
    //-------------------------------------------------------------------------
    size = sub/2;
    // printf("MAP spot (%d %d)==>(%d %d) size %d\n", x0, y0, x0-w, y0-w, size);
    gdk_draw_arc(MAP_PIX, gc, TRUE,
            x0-size-1, y0-size-1, sub, sub, 0, 360 * DEGREE_64);
#endif

}

GtkLabel* map_scale_label = 0;


static void meter_scale_reset()
{
    scale_index = 8;
    pixels_per_meter = scales[scale_index].scale;
}

#if 0
static double meter_scale()
{
    return scales[scale_index].scale;
}
#endif

//------------------------------------------------------------------------------------
//
//      MAP_redraw()
//
//------------------------------------------------------------------------------------

static void
MAP_redraw()
{
    // printf("MAP_redraw\n");
    if (!MAP_PIX) return;
    page_MAP_alloc_colors();

    MAP_clear();
    gtk_button_set_label(button_style, TrackStyle::getLabel());

    computePixelBaseline();
    MAP_draw_all_points();
    gtk_widget_queue_draw_area(GTK_WIDGET(MAP), 0, 0, 476, 476);
    map_have_new_pos = false;

    set_fix_label();
}

//------------------------------------------------------------------------------------
//
//      page_MAP_done()
//
//------------------------------------------------------------------------------------

void
page_MAP_done()
{
    if (!map_have_new_pos) return;
    MAP_redraw();
}

//------------------------------------------------------------------------------------
//
//      cb_meter_scale_change()
//
//  GTK+ callback when the user pressed [+] or [-].
//
//------------------------------------------------------------------------------------

static void
cb_meter_scale_change(GtkWidget* w, gpointer data)
{
    int x = GPOINTER_TO_INT(data);
    int new_index = scale_index + x;
    if (new_index < 0) new_index = 0;
    if (new_index >= (int) MAX_SCALES) new_index = MAX_SCALES - 1;
    if (new_index == scale_index)
    {
        return;         //  nothing changed!
    }
    scale_index = new_index;

    MeterScale& ms = scales[scale_index];
    pixels_per_meter = ms.scale;

    char buf[64];
    sprintf(buf, "Scale %s", ms.name);
    gtk_label_set_text(map_scale_label, buf);
    MAP_recompute_xy_pixels("scale change");
    MAP_redraw();
}

//------------------------------------------------------------------------------------
//
//      MAP_recompute_xy_pixels()
//
//   Since the multiplier changed, we need to recompute the
//   int x,y = (int) double mult*(double meters_x,double meters_y).
//
//------------------------------------------------------------------------------------

static void
MAP_recompute_xy_pixels(const char* msg)
{
    //  Only compute pixel baseline if:
    //  ===>  mult changes  <===
    //  ... or if average changes (when centering the average),
    //  ... or if newer position  (when centering the last position)
    computePixelBaseline();

    lla* p = lla_begin();
    if (p) p->compute_xy_pixels(msg);
    do
    {
        p = lla_next();
        p->compute_xy_pixels(msg);
    } while (!lla_last());
}

//------------------------------------------------------------------------------------
//
//      page_MAP_init()
//
//------------------------------------------------------------------------------------

void
page_MAP_init()
{
    meter_scale_reset();
    memset(LLA, 0, sizeof(LLA));
    nLLA = 0;
    indexLLA = 0;
    centerOfWindow(lla_x0, lla_y0);
    sum_lon = 0.0;
    sum_lat = 0.0;
    lla::restart();
}

//------------------------------------------------------------------------------------
//
//      page_MAP_create()
//
//------------------------------------------------------------------------------------

GtkWidget*
page_MAP_create(GtkWidget* window)
{
    GtkWidget* label;
    GtkWidget* button;
    GtkWidget* w;
    GtkWidget* box;
    GtkObject* adj;
    GtkWidget* table;

    //--------------------------------------------------------------------------
    //          table
    //--------------------------------------------------------------------------
    table = gtk_table_new(14, 2, TRUE);

    //--------------------------------------------------------------------------
    //          label
    //--------------------------------------------------------------------------
    label = gtk_label_new("Position Plot");
    gtk_label_set_justify((GtkLabel*)label, GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE (table), label, 0, 2, 0, 1);
    gtk_widget_show(label);
    map_label = GTK_LABEL(label);

    MAP_text = gtk_widget_create_pango_layout(label, "");

    //--------------------------------------------------------------------------
    //          window for the position plot
    //--------------------------------------------------------------------------
    w  = gtk_drawing_area_new();
    MAP = GTK_DRAWING_AREA(w);
    gtk_drawing_area_size(MAP, 500, 500);

    gtk_signal_connect(GTK_OBJECT(w), "configure_event",
                    G_CALLBACK(configure_MAP_event), w);

    gtk_signal_connect(GTK_OBJECT(w), "expose_event",
                    G_CALLBACK(expose_MAP_event), w);

    gtk_widget_set_events(w, GDK_EXPOSURE_MASK);
    gtk_table_attach_defaults(GTK_TABLE (table), w, 0, 2, 1, 12);
    gtk_widget_show(w);

    //--------------------------------------------------------------------------
    //          Meter scaling
    //--------------------------------------------------------------------------

    box = gtk_hbox_new(FALSE, 1);
    gtk_container_set_border_width (GTK_CONTAINER (box), 0);

    label = gtk_label_new ("Scale x1");
    gtk_widget_show (label);
    map_scale_label = GTK_LABEL(label);
    gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);

    button = gtk_button_new_with_label("+");
    g_signal_connect (G_OBJECT (button), "clicked",
                      G_CALLBACK (cb_meter_scale_change), (gpointer) -1);
    gtk_box_pack_start (GTK_BOX (box), button, TRUE, TRUE, 0);
    gtk_widget_show(button);

    button = gtk_button_new_with_label("-");
    g_signal_connect (G_OBJECT (button), "clicked",
                      G_CALLBACK (cb_meter_scale_change), (gpointer) +1);
    gtk_box_pack_start (GTK_BOX (box), button, TRUE, TRUE, 0);
    gtk_widget_show(button);

    gtk_widget_show (box);

    gtk_table_attach_defaults(GTK_TABLE (table), box, 0, 2, 12, 13);

    //--------------------------------------------------------------------------
    //          Box to hold the center style button
    //          and the position slider.
    //--------------------------------------------------------------------------

    box = gtk_hbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 0);

    //--------------------------------------------------------------------------
    //          Track style button
    //--------------------------------------------------------------------------
    button = gtk_button_new_with_label(TrackStyle::getLabel());
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(on_clicked_track_style), button);
    gtk_box_pack_start (GTK_BOX (box), button, FALSE, TRUE, 0);
    gtk_widget_show(button);
    button_style = GTK_BUTTON(button);

    //--------------------------------------------------------------------------
    //          Position adjuster
    //--------------------------------------------------------------------------
    adj = gtk_adjustment_new(1.0, 1.0, 1.0, 1.0, 1.0, 1.0);
    g_signal_connect (G_OBJECT (adj), "value_changed",
                      G_CALLBACK (cb_change_center), (gpointer) adj);
    adj_center = GTK_ADJUSTMENT(adj);
    scale_center = gtk_hscale_new(adj_center);
    scale_init(GTK_SCALE(scale_center));
    gtk_widget_set_sensitive(scale_center, false);
    gtk_box_pack_start (GTK_BOX (box), scale_center, TRUE, TRUE, 1);
    gtk_widget_show (scale_center);

    gtk_table_attach_defaults(GTK_TABLE (table), box, 0, 1, 13, 14);
    gtk_widget_show (box);

    //--------------------------------------------------------------------------
    //          km/h or mph
    //--------------------------------------------------------------------------
    button = gtk_button_new_with_label("mph");
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(on_clicked_mph), button);
    gtk_table_attach_defaults(GTK_TABLE (table), button, 1, 2, 13, 14);
    gtk_widget_show(button);

    gtk_widget_show(table);

    page_MAP_init();
    return table;
}

static lla t;

static void
page_map_report_dist(FILE* f, const char* msg, lla* p, lla* avg)
{
    fprintf(f, "%s position %f  %f", msg, p->lon, p->lat);

    if (avg != p)
    {
        t = *p;
        t.compute_xy_meters_abs(avg);
        double d = sqrt(t.meters_x*t.meters_x + t.meters_y*t.meters_y);
        if (d > 0.09) fprintf(f, " %.1f m from average", d);
    }
    fprintf(f, "\n");
}

void
page_map_report(FILE* f)
{
    int n = fix_good;
    if (n > MAX_LLA) n = MAX_LLA;
    fprintf(f, "%d positions obtained, %d plotted\n", fix_good, n);

    lla* p0 = lla_begin();
    lla* pA = TrackStyle::getReference(TrackStyle::AVERAGE);
    lla* pN = TrackStyle::getReference(TrackStyle::POINT_LAST);

    page_map_report_dist(f, "  First",  p0, pA);
    page_map_report_dist(f, "Average",  pA, pA);
    page_map_report_dist(f, "   Last",  pN, pA);
}
